const config = require('config');
const fs = require('fs');
const BizResult = require('./BizResult')
const BizResultCode = require('./BaseResultCode');
const logger = require('./logger')('FileController.js');
const fileService = require('./FileService');
const orderService = require('./OrderService');
const zlib = require('zlib');
const fileConfig = config.get('fileConfig');
const uploadFileBufferSize = fileConfig.get('uploadFileBufferSize');
const uploadFileBuffer = Buffer.alloc(uploadFileBufferSize);
const merkleBufferSize = fileConfig.get('merkleBufferSize');
const merkleBuffer = Buffer.alloc(merkleBufferSize);
const crypto = require('crypto');

class FileController {

    /**
     * save file info, contains order_id,file_path,email,file_size,md5,device_type
     * @param {*} req       HTTP request
     * @param {*} res       HTTP response
     * @returns 
     */
    static async save(req, res) {

        var orderId = req.body.orderId;
        var email = req.body.email;
        var filePath = req.body.filePath;
        var fileSize = req.body.fileSize;
        var md5 = req.body.md5;

        //  1 foggie 2 foggieMax 3 client
        var deviceType = req.body.deviceType;

        if (!orderId || !email || !filePath || !fileSize || !md5 || !deviceType) {
            res.send(BizResult.validateFailed(orderId));
            return;
        }

        orderService.updateOrderUsedSpace(email, orderId, fileSize);

        var codebooks = await fileService.getFileCodeBookByMd5(orderId, email, md5);
        if (codebooks instanceof BizResultCode) {
            res.send(BizResult.fail(codebooks));
            return;
        }

        if (codebooks.length == 0) {
            res.send(BizResult.fail(BizResultCode.GET_FILE_CODEBOOK_FAILED));
            return;
        }

        var codebook = codebooks[0];

        fileService.saveFileProp(orderId, email, filePath, fileSize, md5, codebook.cid, deviceType);

        var fileUploadRecordRes = await fileService.getFileUploadRecord(orderId, email, md5);

        if (fileUploadRecordRes instanceof BizResultCode) {
            res.send(BizResult.fail(fileUploadRecordRes));
            return;
        }

        try {
            for (const file of fileUploadRecordRes) {
                const filename = file.file_path;
                fs.unlinkSync(filename);
                logger.info("delete file, file path:{}", filename);
            }
        } catch (error) {
            logger.error("delete file error:{}", error);
        }

        res.send(BizResult.success());
    }

    static async saveFileCodebook(req, res) {
        var orderId = req.body.orderId;
        var email = req.body.email;
        var md5 = req.body.md5;
        var cid = req.body.cid;
        var filePath = req.body.filePath;
        var fileSize = req.body.fileSize;
        var partArr = req.body.partArr;
        var deviceType = req.body.deviceType;
        if (!orderId || !email || !md5 || !cid || !filePath || !fileSize || !partArr || !deviceType) {
            res.send(BizResult.validateFailed());
            return;
        }
        for (const part of partArr) {
            var resultData = await fileService.saveFileCodeBook4ThirdParty(orderId, email, md5, parseInt(fileSize), cid, part.partId, part.base64data, part.hashVaule, 2);
            if (resultData instanceof BizResultCode) {
                res.send(BizResult.fail(resultData));
                return;
            }
        }
        var saveRes = await fileService.saveFileProp4ThirdParty(orderId, email, filePath, parseInt(fileSize), md5, cid, deviceType, 2);

        if (saveRes instanceof BizResultCode) {
            res.send(BizResult.fail(saveRes));
            return;
        }

        res.send(BizResult.success());
    }


    /**
     * file list
     * @param {*} req HTTP request       
     * @param {*} res HTTP response
     * @returns 
     */
    static async list(req, res) {

        var orderId = req.body.orderId;
        var email = req.body.email;
        var pageSize = req.body.pageSize;
        var pageNo = req.body.pageNo;
        var deviceType = req.body.deviceType;

        if (!orderId || !email || !deviceType) {
            res.send(BizResult.validateFailed(orderId));
            return;
        }

        var limit = 10;
        if (typeof (pageSize) !== "undefined") {
            limit = pageSize
        }

        var skip = 0;
        if (typeof (pageNo) !== "undefined") {
            skip = (pageNo - 1) * limit
        }

        var resultData = await fileService.getFileList(orderId, email, deviceType, skip, limit);
        if (resultData instanceof BizResultCode) {
            res.send(BizResult.fail(resultData));
            return;
        }

        var total = await fileService.getFileCount(orderId, email, deviceType);
        if (total instanceof BizResultCode) {
            res.send(BizResult.fail(total));
            return;
        }

        var result = {};
        result['list'] = resultData;
        result['count'] = total;
        res.send(BizResult.success(result));
    }

    /**
     * file all list
     * @param {*} req HTTP request       
     * @param {*} res HTTP response
     * @returns 
     */
    static async allList(req, res) {


        var orderId = req.body.orderId;
        var email = req.body.email;
        var deviceType = req.body.deviceType;

        if (!orderId || !email || !deviceType) {
            res.send(BizResult.validateFailed(orderId));
            return;
        }

        var resultData = await fileService.getAllFile(orderId, email, deviceType);
        if (resultData instanceof BizResultCode) {
            res.send(BizResult.fail(resultData));
            return;
        }
        res.send(BizResult.success(resultData));
    }

    static async zeroCopyUpload4BigFile(req, res) {

        var filePath = req.body.filePath;
        var destPath = req.body.destPath;
        var md5 = req.body.md5;
        var orderId = req.body.orderId;
        var email = req.body.email;
        var fileType = req.body.fileType;
        logger.info("tuning performance,start upload file,md5:{},timestamp:{}", md5, new Date().getTime());
        if (!filePath || !destPath || !md5 || !orderId || !email) {
            res.send(BizResult.validateFailed());
            return;
        }

        var stats = fs.statSync(filePath);
        var fileSize = stats.size;

        // peerId
        var orderInfo = await orderService.getOrderById(email, orderId);
        if (orderInfo instanceof BizResultCode) {
            res.send(BizResult.fail(orderInfo));
            return;
        }

        var token = orderInfo.upload_file_token;
        var peerId = orderInfo.peer_id;
        var rpc = orderInfo.rpc;
        var foggieId = orderInfo.foggie_id;

        const header = {
            peerId: peerId,
            Id: foggieId,
            token: token
        };
        const request = {
            key: destPath,
            md5: md5,
            contentLength: fileSize,
            contentType: fileType
        };
        const putObjectReq = {
            header: header,
            request: request
        };
        var client = await fileService.getProxGrpcClient(rpc, header);

        client.NewMultipartObject(putObjectReq, async function (err, data) {
            if (err) {
                logger.error('err:', err);
                res.send(BizResult.customFail(BizResultCode.CREATE_FILE_FAILED, err.message));
                return;
            }
            var uploadId = data.uploadId;
            if (!uploadId) {
                res.send(BizResult.customFail(BizResultCode.CREATE_FILE_FAILED, 'uploadId is null'));
                return;
            }

            var totalPart = Math.ceil(fileSize / uploadFileBufferSize);
            var fd = fs.openSync(filePath, 'r');
            for (var i = 0; i < totalPart; i++) {
                await fileService.saveFileUploadRecord(orderId, email, destPath, md5, i);
                fs.readSync(fd, uploadFileBuffer, 0, uploadFileBufferSize, i * uploadFileBufferSize);
                var partMd5 = crypto.createHash('md5').update(uploadFileBuffer).digest('hex');
                // offsetArray.forEach(async (offset) => {
                //     // Determine whether the offset range of the current shard file 
                //     // needs to be recorded in a password book
                //     if (offset >= minOffset && offset < maxOffset) {
                //         fs.readSync(fd, merkleBuffer, 0, merkleBufferSize, offset);
                //         var hashVaule = crypto.createHash('md5').update(merkleBuffer).digest('hex');
                //         var compressedData = zlib.gzipSync(merkleBuffer);
                //         var base64data = Buffer.from(compressedData).toString('base64');
                //         //After successfully uploading the big file's cid, 
                //         //it has not yet been generated. 
                //         //First, an empty string will be recorded and updated in the subsequent submission interface
                //         await fileService.saveFileCodeBook(orderId, email, wholeMd5, wholeFileSize, '', offset / merkleBufferSize, base64data, hashVaule);
                //     }
                // });

                await syncUpload(orderId, email, md5, destPath, partMd5, i, uploadId);
            }
            fs.closeSync(fd);

            var uploadRecordRes = await fileService.getFileUploadRecord(orderId, email, md5);
            if (uploadRecordRes instanceof BizResultCode) {
                res.send(BizResult.fail(uploadRecordRes));
                return;
            }

            var unFinishUpload = uploadRecordRes.some((item) => item.state === 0);
            if (unFinishUpload) {
                res.send(BizResult.fail(BizResultCode.UPLOAD_FILE_UNFINISH));
                return;
            }

            var parts = [];
            for (const uploadRecord of uploadRecordRes) {
                parts.push({ 'partNumber': uploadRecord.part_num+1, 'etag': uploadRecord.etag });
            }

            // complete upload file
            const completeUpload = {
                parts: parts
            };
            const request = {
                key: destPath,
                uploadId: uploadId,
                completeUpload: completeUpload
            };
            const completeMultipartReq = {
                header: header,
                request: request
            };
            client.CompleteMultipart(completeMultipartReq, async function (err2, data2) {
                if (err2) {
                    logger.error('err:', err2);
                    res.send(BizResult.customFail(BizResultCode.COMPLETE_FILE_FAILED, err2.message));
                    return;
                }
                logger.info("tuning performance,end upload file,md5:{},timestamp:{}", md5, new Date().getTime());
                res.send(BizResult.success(data2.cid));
            });
        });


        async function syncUpload(orderId, email, md5, destPath, partMd5, i, uploadId) {
            var request = {
                key: destPath,
                md5: partMd5,
                contentLength: uploadFileBufferSize,
                partId: i + 1,
                uploadId: uploadId
            };
            var putObjectPartReq = {
                header: header,
                request: request
            };
            return new Promise((resolve, reject) => {
                var stream = client.PutObjectPart(async function (err, data) {
                    if (err) {
                        logger.error('err:', err);
                        return;
                    }
                    var num = await fileService.updateUploadRecordEtag(orderId, email, md5, destPath, i, data.etag);
                    resolve(num);
                });
                stream.write({ req: putObjectPartReq });
                stream.write({ chunk: uploadFileBuffer });
                stream.end();
            });
        }
        // const request = {
        //     key: destPath,
        //     md5: md5,
        //     contentLength: fileSize,
        //     partId: 1,
        //     uploadId: uploadId
        // };
        // const putObjectPartReq = {
        //     header: header,
        //     request: request
        // };
        // let stream = client.PutObjectPart(async function (err, data) {
        //     if (err) {
        //         logger.error('err:', err);
        //         res.send(BizResult.customFail(BizResultCode.UPLOAD_FILE_FAILED, err.message));
        //         return;
        //     }
        //     // complete upload file
        //     const completeUpload = {
        //         parts: [{ 'partNumber': 1, 'etag': data.etag }]
        //     };
        //     const request = {
        //         key: destPath,
        //         uploadId: uploadId,
        //         completeUpload: completeUpload
        //     };
        //     const completeMultipartReq = {
        //         header: header,
        //         request: request
        //     };
        //     client.CompleteMultipart(completeMultipartReq, async function (err2, data2) {
        //         if (err2) {
        //             logger.error('err:', err2);
        //             res.send(BizResult.customFail(BizResultCode.COMPLETE_FILE_FAILED, err2.message));
        //             return;
        //         }
        //         logger.info("tuning performance,end upload file,md5:{},timestamp:{}", md5, new Date().getTime());
        //         res.send(BizResult.success(data2.cid));
        //     });
        // });
        // stream.write({ req: putObjectPartReq });
        // const readStream = fs.createReadStream(filePath, { highWaterMark: uploadFileBufferSize });

        // let uploadedBytes = 0;
        // let startTime = Date.now();
        // readStream.on('data', (chunk) => {
        //     uploadedBytes += chunk.length;
        //     const elapsedTime = Date.now() - startTime;
        //     logger.info('elapsedTime:{}', elapsedTime);
        //     const uploadSpeed = uploadedBytes / elapsedTime;
        //     logger.info('uploadSpeed:{}', uploadSpeed);
        //     const uploadProgress = uploadedBytes / fileSize;
        //     logger.info('uploadProgress:{}', uploadProgress);
        //     if (!stream.write({ chunk: chunk })) {
        //         readStream.pause();
        //         stream.once('drain', () => {
        //             // in drain event resume read streaming
        //             readStream.resume();
        //         });
        //     }
        // });
        // readStream.on('end', () => {
        //     stream.end();
        //     return;
        // });

        // stream.on('drain', () => {
        //     readStream.resume();
        // });

        // stream.on('finish', () => {
        //     return;
        // });

        // readStream.on('error', (err) => {
        //     logger.error('err:', err);
        //     res.send(BizResult.fail(BizResultCode.READ_FILE_FAILED));
        //     return;
        // });
    }

    static async uploadValid(req, res) {

        var fileName = req.body.fileName;
        var deviceType = req.body.deviceType;
        var orderId = req.body.orderId;
        var email = req.body.email;
        var foggieToken = req.body.foggieToken;

        if (!orderId || !deviceType) {
            res.send(BizResult.validateFailed(orderId));
            return;
        }

        // valid file name
        if (fileName) {
            if (!email) {
                res.send(BizResult.validateFailed(orderId));
                return;
            }
            if (deviceType == 3) {
                // order
                var queryRes = await fileService.getFileByName(orderId, email, fileName, deviceType);
                if (queryRes instanceof BizResultCode) {
                    res.send(BizResult.fail(queryRes));
                    return;
                }
                if (queryRes.length > 0) {
                    res.send(BizResult.fail(BizResultCode.DUPLICATE_FILE_NAME));
                    return;
                }
            }
            else {
                if (!foggieToken) {
                    res.send(BizResult.validateFailed());
                    return;
                }
                // foggie 
                var order = await orderService.getOrderById(email, orderId);
                if (order instanceof BizResultCode) {
                    res.send(BizResult.fail(order));
                    return;
                }
                var peerId = order.peer_id;
                var rpc = order.rpc;
                var foggieId = order.foggie_id;
                const header = {
                    peerId: peerId,
                    Id: foggieId,
                    token: foggieToken
                };

                var findRes = await fileService.findObjectFromServer(rpc, header, fileName);
                if (findRes instanceof BizResultCode) {
                    res.send(BizResult.fail(findRes));
                    return;
                }
                res.send(BizResult.success(true));
                return;
            }
        }

        if (deviceType == 3) {
            // valid merkle state
            var orderDB = await orderService.getOrderById(email, orderId);
            if (orderDB.merkle_state == 0) {
                logger.info("merkle is not complete, orderId:{}", orderId);
                res.send(BizResult.fail(BizResultCode.ORDER_MERKLE_NOT_COMPLETE));
                return;
            }
            // valid order state
            var orderInfo = orderService.getOrderFromChain(orderId);
            if (orderInfo instanceof BizResultCode) {
                logger.info("get order from chain failed, orderId:{} ", orderId);
                res.send(BizResult.fail(orderInfo));
                return;
            }
            if (orderInfo.state == 4 || orderInfo.state == 5) {
                logger.info("order state is invalid, can't upload file, orderId:{}", orderId);
                res.send(BizResult.fail(BizResultCode.ORDER_STATE_INVALID_UPLOAD));
                return;
            }

            // upload file valid challenge is end
            var challengeList = orderService.getChallengeByState(orderId, [0, 1, 3, 7]);
            if (challengeList instanceof BizResultCode) {
                res.send(BizResult.fail(challengeList));
                return;
            }
            if (challengeList.length > 0) {
                for (const chanllenge of challengeList) {
                    if (chanllenge.state === 3) {
                        logger.info("challenge is not response, orderId:{}", orderId);
                        res.send(BizResult.fail(BizResultCode.ORDER_CHALLENGE_NOT_RESPONSE));
                        return;
                    }
                    if (chanllenge.state === 7) {
                        logger.info("order is end, orderId:{}", orderId);
                        res.send(BizResult.fail(BizResultCode.ORDER_STATE_END));
                        return;
                    }
                    if (chanllenge.state === 0 && chanllenge.pre_merkle_root != chanllenge.merkle_root) {
                        logger.info("merkle is inconsistent, orderId:{}", orderId);
                        res.send(BizResult.fail(BizResultCode.MERKLE_INCONSISTENT));
                        return;
                    }
                    if (chanllenge.state === 1 && parseInt(chanllenge.pre_merkle_root) != 0) {
                        logger.info("merkle is inconsistent, orderId:{}", orderId);
                        res.send(BizResult.fail(BizResultCode.MERKLE_INCONSISTENT));
                        return;
                    }
                }
            }
        }
        res.send(BizResult.success(true));
    }

    /**
     * create file
     * @param {*} req 
     * @param {*} res     
     * @returns 
     */
    static async create(req, res) {

        var fileName = req.body.fileName;
        var md5 = req.body.md5;
        var fileType = req.body.fileType;
        var fileSize = req.body.fileSize;
        var orderId = req.body.orderId;
        var email = req.body.email;
        var deviceType = req.body.deviceType;
        var foggieToken = req.body.foggieToken;

        if (!fileName || !fileSize || !orderId || !md5 || !email || !deviceType) {
            res.send(BizResult.validateFailed());
            return;
        }

        if (deviceType != 3) {
            var foggieTokenArr = foggieToken.split(' ');
            if (foggieTokenArr.length != 2) {
                res.send(BizResult.validateFailed('foggieToken: ' + foggieToken));
                return;
            }
            foggieToken = foggieTokenArr[1];
        }

        // valid the same as file
        var resultData = await fileService.getFileByMd5(orderId, email, md5, deviceType)

        if (resultData instanceof BizResultCode) {
            res.send(BizResult.fail(resultData));
            return;
        }
        if (resultData.length > 0) {
            res.send(BizResult.customFail(BizResultCode.FILE_IS_EXIST, ',conflict filename:' + decodeURIComponent(resultData[0].file_path)));
            return;
        }

        // peerId
        var orderInfo = await orderService.getOrderById(email, orderId);
        if (orderInfo instanceof BizResultCode) {
            res.send(BizResult.fail(orderInfo));
            return;
        }

        var token;
        if (deviceType == 3) {
            // token
            token = orderInfo.upload_file_token;
            // valid used space
            var usedSpace = orderInfo.used_space;
            var totalSpace = orderInfo.total_space;
            if (usedSpace + parseInt(fileSize) > totalSpace) {
                logger.error("used space is over, orderId:{}, usedSpace:{}, fileSize:{} totalSpace:{}", orderId, usedSpace, fileSize, totalSpace);
                res.send(BizResult.fail(BizResultCode.ORDER_REMAINING_SPACE_NOT_ENOUGH));
                return;
            }
        }
        else {
            token = foggieToken;
        }

        var peerId = orderInfo.peer_id;
        var rpc = orderInfo.rpc;
        var foggieId = orderInfo.foggie_id;

        const header = {
            peerId: peerId,
            Id: foggieId,
            token: token
        };
        const request = {
            key: fileName,
            md5: md5,
            contentLength: fileSize,
            contentType: fileType
        };
        const putObjectReq = {
            header: header,
            request: request
        };
        var client = await fileService.getProxGrpcClient(rpc, header);

        client.NewMultipartObject(putObjectReq, function (err, data) {
            if (err) {
                logger.error('err:', err);
                res.send(BizResult.customFail(BizResultCode.CREATE_FILE_FAILED, err.message));
                return;
            }
            var duplicateRes = {};
            duplicateRes['duplicateUpload'] = false;
            duplicateRes['uploadId'] = data.uploadId;
            res.send(BizResult.success(duplicateRes));
        });
    }

    /**
     * upload file
     * 
     * @param {*} req HTTP request
     * @param {*} res HTTP response
     * @returns
     */
    static async upload(req, res) {
        logger.info("tuning performance,start upload file,md5:{},timestamp:{}", md5, new Date().getTime());
        // 1 small 2 big
        var fileCategory = req.body.fileCategory;
        // deviceType 1 foggie 2 foggieMax 3 client
        var deviceType = parseInt(req.body.deviceType);
        // var deviceType = req.body.deviceType;
        var fileName = req.body.fileName;
        var md5 = req.body.md5;
        var fileType = req.body.fileType;
        var fileSize = req.body.fileSize;
        var orderId = req.body.orderId;
        var email = req.body.email;
        // the following are big file parameter
        var partId = req.body.partId;
        var uploadId = req.body.uploadId;
        var minOffset = req.body.minOffset;
        var maxOffset = req.body.maxOffset;
        var wholeMd5 = req.body.wholeMd5;
        var wholeFileSize = req.body.wholeFileSize;
        var foggieToken = req.body.foggieToken;

        if (!fileCategory || !fileName || !md5 || !fileSize || !orderId || !email || !wholeMd5 || !deviceType) {
            res.send(BizResult.validateFailed());
            return;
        }

        if (deviceType != 3) {
            var foggieTokenArr = foggieToken.split(' ');
            if (foggieTokenArr.length != 2) {
                res.send(BizResult.validateFailed('foggieToken: ' + foggieToken));
                return;
            }
            foggieToken = foggieTokenArr[1];
        }

        // check small file is exist
        if (parseInt(fileCategory) == 1) {
            var resultData = await fileService.getFileByMd5(orderId, email, wholeMd5, deviceType)

            if (resultData instanceof BizResultCode) {
                res.send(BizResult.fail(resultData));
                return;
            }
            if (resultData.length > 0) {
                res.send(BizResult.customFail(BizResultCode.FILE_IS_EXIST, ',conflict filename:' + decodeURIComponent(resultData[0].file_path)));
                return;
            }
        }
        // peerId
        var orderFromDB = await orderService.getOrderById(email, orderId);

        if (orderFromDB instanceof BizResultCode) {
            res.send(BizResult.fail(orderFromDB));
            return;
        }
        var token;
        if (deviceType == 3) {
            // token
            token = orderFromDB.upload_file_token;
        }
        else {
            token = foggieToken;
        }
        var peerId = orderFromDB.peer_id;
        var foggieId = orderFromDB.foggie_id;

        const header = {
            peerId: peerId,
            Id: foggieId,
            token: token
        };
        var file = req.files.file;
        var rpc = orderFromDB.rpc;

        //small file upload
        if (parseInt(fileCategory) == 1) {
            await smallFileUpload(fileName, md5, fileSize, fileType, orderFromDB, header, res, fileCategory, orderId, email, peerId, file, foggieId, deviceType);
        }
        // big file upload
        else if (parseInt(fileCategory) == 2) {
            if (!partId || !uploadId || !minOffset || !maxOffset || !wholeFileSize) {
                res.send(BizResult.validateFailed());
                return;
            }

            const request = {
                key: fileName,
                md5: md5,
                contentLength: fileSize,
                partId: partId,
                uploadId: uploadId
            };
            const putObjectPartReq = {
                header: header,
                request: request
            };
            var netClient = await fileService.getProxGrpcClient(rpc, header);

            logger.info("tuning performance,start deal PutObjectPart,md5:{},timestamp:{}", md5, new Date().getTime());
            let stream = netClient.PutObjectPart(async function (err, data) {
                if (err) {
                    logger.error('err:', err);
                    res.send(BizResult.customFail(BizResultCode.UPLOAD_FILE_FAILED, err.message));
                    return;
                }
                logger.info("tuning performance,end deal PutObjectPart,md5:{},timestamp:{}", md5, new Date().getTime());
                // After uploading the sharded file, 
                // it is necessary to record the temporary path of the sharded file 
                // and the MD5 of the entire large file, 
                //which is used to generate a merkle tree in the commit interface
                var saveFileUploadRecordRes = await fileService.saveFileUploadRecord(orderId, email, file.path, wholeMd5, partId)
                if (saveFileUploadRecordRes instanceof BizResultCode) {
                    res.send(BizResult.fail(saveFileUploadRecordRes));
                    return;
                }
                // Calculate the offset array of the password book 
                // based on the size of the file and the block size of 
                // the merkle tree
                var offsetArray = await fileService.getCodebookOffset(fileCategory, orderId, email, fileName, wholeMd5, wholeFileSize, merkleBufferSize, 2)
                if (offsetArray instanceof BizResultCode) {
                    res.send(BizResult.fail(offsetArray));
                    return;
                }

                var fd = fs.openSync(file.path, 'r');
                offsetArray.forEach(async (offset) => {
                    // Determine whether the offset range of the current shard file 
                    // needs to be recorded in a password book
                    if (offset >= minOffset && offset < maxOffset) {
                        fs.readSync(fd, merkleBuffer, 0, merkleBufferSize, offset);
                        var hashVaule = crypto.createHash('md5').update(merkleBuffer).digest('hex');
                        var compressedData = zlib.gzipSync(merkleBuffer);
                        var base64data = Buffer.from(compressedData).toString('base64');
                        //After successfully uploading the big file's cid, 
                        //it has not yet been generated. 
                        //First, an empty string will be recorded and updated in the subsequent submission interface
                        await fileService.saveFileCodeBook(orderId, email, wholeMd5, wholeFileSize, '', offset / merkleBufferSize, base64data, hashVaule);
                    }
                });
                fs.closeSync(fd);
                logger.info("tuning performance,end upload file,md5:{},timestamp:{}", md5, new Date().getTime());
                res.send(BizResult.success(data.etag));
            });

            logger.info("tuning performance,start deal upload steaming,md5:{},timestamp:{}", md5, new Date().getTime());
            stream.write({ req: putObjectPartReq });
            const readStream = fs.createReadStream(file.path, { highWaterMark: uploadFileBufferSize });
            readStream.on('data', (chunk) => {
                logger.info("tuning performance,start stream.write,md5:{},timestamp:{}", md5, new Date().getTime());
                if (!stream.write({ chunk: chunk })) {
                    readStream.pause();
                    stream.once('drain', () => {
                        // in drain event resume read streaming
                        readStream.resume();
                    });
                }
                logger.info("tuning performance,end stream.write,md5:{},timestamp:{}", md5, new Date().getTime());
            });
            readStream.on('end', () => {
                logger.info("tuning performance,end deal upload steaming,md5:{},timestamp:{}", md5, new Date().getTime());
                stream.end();
                return;
            });

            stream.on('drain', () => {
                readStream.resume();
            });

            stream.on('finish', () => {
                return;
            });

            readStream.on('error', (err) => {
                logger.error('err:', err);
                res.send(BizResult.fail(BizResultCode.READ_FILE_FAILED));
                return;
            });
        }
        else {
            res.send(BizResult.fail(BizResultCode.NOT_SUPPORT_FILE_TYPE));
        }
    }

    /**
     * complete upload file
     * 
     * @param {*} req HTTP request
     * @param {*} res HTTP response
     * @returns
     */
    static async complete(req, res) {

        var fileName = req.body.fileName;
        var orderId = req.body.orderId;
        var parts = req.body.parts;
        var uploadId = req.body.uploadId;
        var md5 = req.body.md5;
        var fileSize = req.body.fileSize;
        var email = req.body.email;
        var deviceType = req.body.deviceType;
        var foggieToken = req.body.foggieToken;

        if (!fileName || !orderId || !parts || !uploadId || !md5 || !fileSize || !email || !deviceType) {
            res.send(BizResult.validateFailed());
            return;
        }

        if (deviceType != 3) {
            var foggieTokenArr = foggieToken.split(' ');
            if (foggieTokenArr.length != 2) {
                res.send(BizResult.validateFailed('foggieToken: ' + foggieToken));
                return;
            }
            foggieToken = foggieTokenArr[1];
        }

        // peerId
        var orderInfo = await orderService.getOrderById(email, orderId);
        if (orderInfo instanceof BizResultCode) {
            res.send(BizResult.fail(orderInfo));
            return;
        }

        var token;
        if (deviceType == 3) {
            // token
            token = orderInfo.upload_file_token;
        }
        else {
            token = foggieToken;
        }

        var peerId = orderInfo.peer_id;
        var rpc = orderInfo.rpc;
        var foggieId = orderInfo.foggie_id;

        // According to the upload record of the file, reread the file once
        var fileUploadRecordRes = await fileService.getFileUploadRecord(orderId, email, md5);

        if (fileUploadRecordRes instanceof BizResultCode) {
            res.send(BizResult.fail(fileUploadRecordRes));
            return;
        }

        // Generate Merkle Tree and Submit
        const header = {
            peerId: peerId,
            Id: foggieId,
            token: token
        };
        const completeUpload = {
            parts: parts
        };
        const request = {
            key: fileName,
            uploadId: uploadId,
            completeUpload: completeUpload
        };
        const completeMultipartReq = {
            header: header,
            request: request
        };
        var netClient = await fileService.getProxGrpcClient(rpc, header);
        netClient.CompleteMultipart(completeMultipartReq, async function (err2, data2) {
            if (err2) {
                logger.error('err:', err2);
                res.send(BizResult.customFail(BizResultCode.COMPLETE_FILE_FAILED, err2.message));
                return;
            }

            logger.info('complete upload success, orderId:{}, data:{}', orderId, data2);

            if (deviceType == 3) {
                // generate function 
                var buildMerkleResult = await buildMerkleTree();
                if (buildMerkleResult.errCode instanceof BizResultCode) {
                    res.send(BizResult.customFail(buildMerkleResult.errCode, buildMerkleResult.errMsg));
                    return;
                }
                res.send(BizResult.success(data2.cid));
            }
            else {
                res.send(BizResult.success(data2.cid));
            }
        });


        async function buildMerkleTree() {
            var powClient = fileService.getPowGrpcClient();
            return new Promise((resolve, reject) => {
                var merkleStream = powClient.BuildMerkelLeaf(function (err, data) {
                    if (err) {
                        logger.error('err:', err);
                        var errCode = BizResultCode.BUILD_MERKLE_FAILED;
                        var errMsg = err.message
                        resolve({ errCode, errMsg });
                        return;
                    }
                    logger.info('generate merkle success, orderId:{}, data:{}', orderId, data);
                    // Update cid after generating merkle tree
                    var updateCodebookRes = fileService.updateFileCodeBook(orderId, email, md5, data.cid);

                    if (updateCodebookRes instanceof BizResultCode) {
                        var errCode = updateCodebookRes;
                        var errMsg = updateCodebookRes.desc;
                        resolve({ errCode, errMsg });
                        return;
                    }
                    resolve(data);
                });
                const mkHeader = {
                    peerId: peerId,
                    Id: foggieId
                };
                const putObjectMKRequest = {
                    key: fileName,
                    size: fileSize
                };
                const putObjectMKReq = {
                    header: mkHeader,
                    request: putObjectMKRequest
                };
                merkleStream.write({ req: putObjectMKReq });

                try {
                    for (const file of fileUploadRecordRes) {
                        var data = fs.readFileSync(file.file_path);
                        merkleStream.write({ chunk: data });
                    }
                }
                catch (err) {
                    logger.error('read file error, fileUploadRecordRes:{}, err: {}', fileUploadRecordRes, err);
                    var errCode = BizResultCode.READ_FILE_FAILED;
                    var errMsg = err.message;
                    resolve({ errCode, errMsg });
                    return;
                }
                merkleStream.end();
            });
        }
    }

    /**
     * query file by name or cid
     * @param {*} req 
     * @param {*} res 
     */
    static async query(req, res) {
        var orderId = req.body.orderId;
        var email = req.body.email;
        var deviceType = req.body.deviceType;
        var key = req.body.key;
        if (!email || !orderId || !deviceType || !key) {
            res.send(BizResult.validateFailed());
            return;
        }
        var resultData = await fileService.queryFile(email, orderId, deviceType, key);
        if (resultData instanceof BizResultCode) {
            res.send(BizResult.fail(resultData));
            return;
        }
        res.send(BizResult.success(resultData));
    }

    /**
     * Delete the generated password book offset array
     * @param {*} req 
     * @param {*} res 
     * @returns 
     */
    static async deleteFileCodebookOffset(req, res) {
        var email = req.body.email;
        var orderId = req.body.orderId;
        var md5 = req.body.md5;

        if (!orderId || !md5 || !email) {
            res.send(BizResult.validateFailed());
            return;
        }

        var resultData = await fileService.deleteCodebookOffset(orderId, email, md5)
        if (resultData instanceof BizResultCode) {
            res.send(BizResult.fail(resultData));
            return;
        }

        res.send(BizResult.success());
    }

    static async deleteFileProp(req, res) {
        var orderId = req.body.orderId;
        var email = req.body.email;
        var md5 = req.body.md5;
        var deviceType = req.body.deviceType;
        var resultData = await fileService.deleteFileByMd5(orderId, email, deviceType, md5)

        if (resultData instanceof BizResultCode) {
            res.send(BizResult.fail(resultData));
            return;
        }
        res.send(BizResult.success());
    }

    /**
     * get code book
     * @param {*} req 
     * @param {*} res 
     * @returns 
     */
    static async getCodebook(req, res) {
        var orderId = req.body.orderId;
        var email = req.body.email;
        var md5 = req.body.md5;
        if (!orderId || !email || !md5) {
            res.send(BizResult.validateFailed());
            return;
        }
        var codebook = await fileService.getCodeBookByOrderId(orderId, email, md5);
        if (codebook instanceof BizResultCode) {
            res.send(BizResult.fail(codebook));
            return;
        }
        res.send(BizResult.success(codebook));
    }

    /**
     * reload code book
     * @param {*} req 
     * @param {*} res 
     * @returns 
     */
    static async reloadCodebook(req, res) {
        var codebook = await fileService.reloadCodeBook();
        if (codebook instanceof BizResultCode) {
            res.send(BizResult.fail(codebook));
            return;
        }
        res.send(BizResult.success(codebook));
    }

}
module.exports = FileController

async function smallFileUpload(fileName, md5, fileSize, fileType, orderFromDB, header, res, fileCategory, orderId, email, peerId, file, foggieId, deviceType) {

    if (deviceType == 3) {
        // valid used space
        var usedSpace = orderFromDB.used_space;
        var totalSpace = orderFromDB.total_space;
        if (usedSpace + parseInt(fileSize) > totalSpace) {
            logger.error("used space is over, orderId:{}, usedSpace:{}, fileSize:{} totalSpace:{}", orderId, usedSpace, fileSize, totalSpace);
            res.send(BizResult.fail(BizResultCode.ORDER_REMAINING_SPACE_NOT_ENOUGH));
            return;
        }
    }
    var rpc = orderFromDB.rpc;
    var netClient = await fileService.getProxGrpcClient(rpc, header);
    const request = {
        key: fileName,
        md5: md5,
        contentLength: fileSize,
        contentType: fileType
    };
    const putObjectReq = {
        header: header,
        request: request
    };

    let uploadFileStream = netClient.PutObject(async function (err, data) {
        if (err) {
            logger.error('err:', err);
            res.send(BizResult.fail(BizResultCode.UPLOAD_FILE_FAILED));
            return;
        }
        // After successful upload, save the file upload record. 
        // The small file does not have shards, so partNum is 0
        fileService.saveFileUploadRecord(orderId, email, file.path, md5, 0)
        var offsetArray = await fileService.getCodebookOffset(fileCategory, orderId, email, fileName, md5, fileSize, merkleBufferSize, 2);
        if (offsetArray instanceof BizResultCode) {
            res.send(BizResult.fail(offsetArray));
            return;
        }
        if (deviceType == 3) {
            var powClient = fileService.getPowGrpcClient();
            let merkleStream = powClient.BuildMerkelLeaf(function (err, data2) {
                if (err) {
                    logger.error('err:', err);
                    res.send(BizResult.customFail(BizResultCode.BUILD_MERKLE_FAILED, " err :" + err.message));
                    return;
                }
                res.send(BizResult.success(data.cid));
                return;
            });
            const mkHeader = {
                peerId: peerId,
                Id: foggieId
            };
            const putObjectMKRequest = {
                key: fileName,
                cid: data.cid,
                size: fileSize
            };
            const putObjectMKReq = {
                header: mkHeader,
                request: putObjectMKRequest
            };
            merkleStream.write({ req: putObjectMKReq });

            var fd = fs.openSync(file.path, 'r');

            for (const offset of offsetArray) {
                fs.readSync(fd, merkleBuffer, 0, merkleBufferSize, offset);
                var hashVaule = crypto.createHash('md5').update(merkleBuffer).digest('hex');
                var compressedData = zlib.gzipSync(merkleBuffer);
                var base64data = Buffer.from(compressedData).toString('base64');
                fileService.saveFileCodeBook(orderId, email, md5, fileSize, data.cid, offset / merkleBufferSize, base64data, hashVaule);
            }
            fs.closeSync(fd);
            merkleStream.write({ chunk: fs.readFileSync(file.path) });
            merkleStream.end();
        }
        else {
            res.send(BizResult.success(data.cid));
        }
    });

    uploadFileStream.write({ req: putObjectReq });

    const readStream = fs.createReadStream(file.path, { highWaterMark: uploadFileBufferSize });

    readStream.on('data', (chunk) => {
        uploadFileStream.write({ chunk: chunk });
    });

    readStream.on('end', () => {
        uploadFileStream.end();
        return;
    });

    readStream.on('error', (err) => {
        logger.error('err:', err);
        res.send(BizResult.customFail(BizResultCode.READ_FILE_FAILED, " err:" + err.message));
        return;
    });
    return netClient;
}
